import requests
import datetime
import ee
import os
import importlib
import sys
# from io import StringIO

class EarthEngine:
    """


    pyeogpr.EarthEngine
    --------------------

        projectID : string
                    Your GEE projectID. Usually starts with "ee-...". You can find it next to your
                    profile picture, top right corner on: https://code.earthengine.google.com

        sensor :    string
            Satellite sensor to use.
            You can search for it on: https://developers.google.com/earth-engine/datasets/catalog
            Insert the string, as you would find it on the GEE catalog site:
            for example: "COPERNICUS/S3/OLCI" or "LANDSAT/LC08/C02/T1_L2"

        biovar :    string
            Biophysical variable to process.

            Currently "built-in" variables available for each sensor:

            - SENTINEL2_L1C: Cab, Cm, Cw, FVC, LAI, laiCab, laiCm, laiCw
            - SENTINEL2_L2A: Cab, Cm, Cw, FVC, LAI, laiCab, laiCm, laiCw, CNC_Cab, CNC_Cprot
            - SENTINEL3_OLCI_L1B: FAPAR, FVC, LAI, LCC

            If you have your own model trained with ARTMO (https://artmotoolbox.com), you need to
            insert the directory of the model, for example: r"C:/User/Models/My_custom_model.py"
            These models need to be in ".py" format, in order to achieve it, please consult:
            https://github.com/SentiFLEXinel/ARTMOtoGEE/

        bounding_box : list, ee.assetpath
            Your region of interest. Insert bbox as list. Can be selected from https://geojson.io/
            (e.g.: [-4.55, 42.73,-4.48, 42.77]).
            Alternatively, you can insert shapefiles, that are already uploaded to your GEE assets.
            Just copy the ee.assetpath which stores your SHP shapefile that you already uploaded to GEE.

        temporal_extent : list
            Your temporal extent to be processed. (e.g.: ["2021-01-01", "2021-12-31"])

        spatial_resolution : int
            Spatial resolution of the exported data. Value in meters.

        cloudmask : Boolean
            If set to "True", cloud masking will be done for Sentinel 2 and 3 sensors. Defaults to False.


    """

    def __init__(
        self,
        projectID,
        sensor,
        biovar,
        bounding_box,
        temporal_extent,
        spatial_resolution,
        cloudmask=False,
    ):
        self.projectID = projectID
        # EE Auth
        ee.Authenticate()
        ee.Initialize(project=self.projectID)

        self.sensor = sensor
        if self.sensor == "SENTINEL2_L1C" or "COPERNICUS/S2_HARMONIZED":
            self.sensor = "COPERNICUS/S2_HARMONIZED"
            search_sensor = "SENTINEL2_L1C"

        if self.sensor == "SENTINEL2_L2A" or "COPERNICUS/S2_SR_HARMONIZED":
            self.sensor = "COPERNICUS/S2_SR_HARMONIZED"
            search_sensor = "SENTINEL2_L2A"

        if self.sensor == "SENTINEL3_L1B" or "COPERNICUS/S3/OLCI":
            self.sensor = "COPERNICUS/S3/OLCI"
            search_sensor = "SENTINEL3_L1B"

        if biovar[-3:] == ".py":
            self.custom_model = biovar
            self.biovar = biovar.split("\\")[-1].split(".")[0]

        else:
            self.biovar = biovar

        if type(bounding_box) is list:  # To accept GeoJSON bbox
            self.bbox = ee.Geometry.BBox(
                bounding_box[0], bounding_box[1], bounding_box[2], bounding_box[3]
            )
        if "projects/ee-" in bounding_box:  # So it accepts shps too!
            self.bbox = ee.FeatureCollection(bounding_box)
            # self.bbox_fc2geo = ee.FeatureCollection(bounding_box).geometry()
            
        self.temporal_extent = temporal_extent
        self.spatial_resolution = spatial_resolution
        self.cloudmask = cloudmask
        self.timeWindows = None
        self.assetpath = None
        self.gpr_model = None
        self.reducer = None

        # User defined or default models import
        if biovar[-3:] == ".py":
            self.custom_model = biovar
            spec = importlib.util.spec_from_file_location(biovar, self.custom_model)
            user_module = importlib.util.module_from_spec(spec)
            sys.modules["user_module"] = user_module
            spec.loader.exec_module(user_module)

            self.model_imported = user_module
            print(self.model_imported.hyp_sig0_GREEN)

        else:
            url = f"https://raw.githubusercontent.com/daviddkovacs/pyeogpr/refs/heads/main/models/{search_sensor}_{self.biovar}_GEE.py"
            print(url)
            response = requests.get(url)

            with open("model_imported.py", "w") as f:
                f.write(response.text)

            if "model_imported" in sys.modules:
                del sys.modules["model_imported"]  # We do this to flush existing modules
            import model_imported

            self.model_imported = model_imported
            print(model_imported.hyp_sig0_GREEN)

        self.imcol = ee.ImageCollection(self.sensor).filterDate(
            ee.Date(self.temporal_extent[0]), ee.Date(self.temporal_extent[1])
        )

    def construct_datacube(self, composite=None):
        """

        composite :  Defaults to None
            The algorithm creates temporal composites, according to the number of days you assign.
            The temporal composites need to be 1 day smaller than the defined temporal_extent.
        """
        self.timeWindows = composite
        print(f"Temporal composites: {self.timeWindows} days")

        dif = (
            datetime.datetime.strptime(self.temporal_extent[1], "%Y-%m-%d").toordinal()
            - datetime.datetime.strptime(
                self.temporal_extent[0], "%Y-%m-%d"
            ).toordinal()
        )
        if dif <= composite:
            raise ValueError(
                f"Temporal composite ({str(composite)} days)  has to be shorter than defined temporal_extent"
            )

    def process_map(self, assetpath=None):
        """

        assetpath: str
            You need to define, which GEE asset (ImageCollection) you would want the maps to be exported to.
            To create your asset, go to top left corner on: https://code.earthengine.google.com/
            You will find three tabs: Scripts, Docs and Assets. Go to Assets and create a new ImageCollection.
            When you created your ImageCollection, copy its ID as a string and assign it to assetpath. If left blank, the script will automatically generate you a default asset.

        """

        if assetpath == None:
            assetpath = f"projects/{self.projectID}/assets/PyEOGPR_{self.biovar}"
            ee.data.createAsset({"type": "ImageCollection"}, assetpath)
            print(f"Assetpath NOT DEFINED:\nExporting to default EE asset: {assetpath}")
        else:
            print(f"Exporting to {assetpath}")
        self.assetpath = assetpath
        self.maploop()

    def sequence_GREEN(self, variable):
        # print(variable)  # Later this probably needs to be changed to select model
        sequence_GREEN = []
        model = self.model_imported
        for i in range(0, model.XTrain_dim_GREEN):
            sequence_GREEN.append(str(i))
        return sequence_GREEN

    def getInputDates(self, i):
        fecha_inicio = ee.Date(self.temporal_extent[0]).advance(
            ee.Number(i).multiply(self.timeWindows), "day"
        )
        fecha_fin = fecha_inicio.advance(self.timeWindows, "day")
        # fecha_fin = endDateGEE()
        fecha_str = datetime.datetime.utcfromtimestamp(
            fecha_inicio.getInfo()["value"] / 1000.0
        ).strftime("%Y%m%d")
        return {
            "fecha_inicio": fecha_inicio,
            "fecha_fin": fecha_fin,
            "fecha_str": fecha_str,
        }

    def maskS3badPixels(self, image):
        qa = ee.Image(image.select("quality_flags"))
        coastLine = 1 << 30
        inLandWater = 1 << 29
        bright = 1 << 27
        invalid = 1 << 25
        Oa12Sat = 1 << 9
        mask = (
            qa.bitwiseAnd(coastLine)
            .eq(0)
            .And(qa.bitwiseAnd(inLandWater).eq(0))
            .And(qa.bitwiseAnd(bright).eq(0))
        )
        return image.updateMask(mask)

    def addVariables(image):
        date = ee.Date(image.get("system:time_start"))
        years = date.difference(ee.Date("1970-01-01"), "days")
        return image.addBands(ee.Image(years).rename("t").int())

    def calculate_GREEN(self, fecha_inicio, fecha_fin, variable):
        model = self.model_imported
        image = ee.Image(
            self.imcol.filterDate(fecha_inicio, fecha_fin)
            .filterBounds(ee.FeatureCollection(self.bbox))
            .map(self.maskS3badPixels)
            .select(model.bands)  # .cast(model.bands_dict, model.bands)
            .max()
            .clipToCollection(ee.FeatureCollection(self.bbox))
        )
        # .clip(fc));
        # TODO: I should play around with this .clip() here..
        im_norm_ell2D_hypell = (
            image.subtract(model.mx_GREEN)
            .divide(model.sx_GREEN)
            .multiply(model.hyp_ell_GREEN)
            .toArray()
            .toArray(1)
        )
        im_norm_ell2D = (
            image.subtract(model.mx_GREEN).divide(model.sx_GREEN).toArray().toArray(1)
        )
        PtTPt = (
            im_norm_ell2D_hypell.matrixTranspose()
            .matrixMultiply(im_norm_ell2D)
            .arrayProject([0])
            .multiply(-0.5)
        )

        PtTDX = (
            ee.Image(model.X_train_GREEN)
            .matrixMultiply(im_norm_ell2D_hypell)
            .arrayProject([0])
            .arrayFlatten([self.sequence_GREEN(self.biovar)])
        )
        arg1 = PtTPt.exp().multiply(model.hyp_sig0_GREEN)

        k_star = PtTDX.subtract(model.XDX_pre_calc_GREEN.multiply(0.5)).exp().toArray()
        mean_pred = k_star.arrayDotProduct(
            model.alpha_coefficients_GREEN.toArray()
        ).multiply(arg1)
        mean_pred = mean_pred.toArray(1).arrayProject([0]).arrayFlatten([[self.biovar]])
        mean_pred = mean_pred.add(model.mean_model_GREEN)
        filterDown = mean_pred.gt(0)

        mean_pred = mean_pred.multiply(filterDown)

        k_star_uncert = (
            PtTDX.subtract(model.XDX_pre_calc_GREEN.multiply(0.5))
            .exp()
            .multiply(arg1)
            .toArray()
        )
        Vvector = (
            ee.Image(model.LMatrixInverse)
            .matrixMultiply(k_star_uncert.toArray(0).toArray(1))
            .arrayProject([0])
        )
        Variance = (
            ee.Image(model.hyp_sig_GREEN)
            .toArray()
            .subtract(Vvector.arrayDotProduct(Vvector))
            .abs()
            .sqrt()
        )
        Variance = (
            Variance.toArray(1)
            .arrayProject([0])
            .arrayFlatten([[self.biovar + "_UNCERTAINTY"]])
        )

        image = image.addBands(mean_pred)
        image = image.addBands(Variance)

        return image.select(self.biovar, self.biovar + "_UNCERTAINTY")

    def maploop(self):
        startDate = datetime.datetime.strptime(
            self.temporal_extent[0], "%Y-%m-%d"
        ).date()
        daysIterations = abs(
            (
                startDate
                - datetime.datetime.strptime(self.temporal_extent[1], "%Y-%m-%d").date()
            )
            // self.timeWindows
        ).days

        for i in range(0, daysIterations):
            print(self.getInputDates(i)["fecha_str"])

            imageHolder = ee.Image().set(
                "system:time_start",
                ee.Date(self.temporal_extent[0])
                .advance(ee.Number(i).multiply(self.timeWindows), "days")
                .millis(),
            )

            imagen = self.calculate_GREEN(
                self.getInputDates(i)["fecha_inicio"],
                self.getInputDates(i)["fecha_fin"],
                self.biovar,
            )  # .multiply(10000).toInt32()
            imageHolder = imageHolder.addBands(imagen)
            image_export = imageHolder.select(self.biovar, self.biovar + "_UNCERTAINTY")

            # Optional masking
            bare_cover = (
                ee.Image("COPERNICUS/Landcover/100m/Proba-V-C3/Global/2019")
                .select("bare-coverfraction")
                .lte(90)
            )
            lakes = (
                ee.Image("COPERNICUS/Landcover/100m/Proba-V-C3/Global/2019")
                .select("discrete_classification")
                .eq(80)
            )
            lakemask = lakes.eq(0)
            # image_export = image_export.mask(lakemask)
            # image_export = image_export.mask(bare_cover)
            
            # SHP-bbox misery
            region = None
            if type(self.bbox) is list:
                region=self.bbox
                image_export = image_export.clip(region)
            elif isinstance(self.bbox, ee.FeatureCollection):
                region = self.bbox.geometry()
                image_export = image_export.clip(region)
            
            image_export = image_export.set(
                "system:time_start",
                ee.Date(self.temporal_extent[0])
                .advance(ee.Number(i).multiply(self.timeWindows), "days")
                .millis(),
            )

            # Export the image to an asset
            exportar = ee.batch.Export.image.toAsset(
                assetId=self.assetpath
                + "/"
                + self.getInputDates(i)["fecha_str"]
                + "_"
                + self.biovar,
                image=image_export,
                maxPixels=17210617060,
                description=self.getInputDates(i)["fecha_str"] + "_" + self.biovar,
                scale=self.spatial_resolution,
                region = region,  # .bounds().getInfo()['coordinates']
            )
            exportar.start()
            exportar.status()



import openeo
import numpy as np
import scipy
import scipy.signal
import ee

import importlib.util
from pyeogpr.sensors import sensors_dict
from pyeogpr.udfgpr import udf_gpr, custom_model_import
from pyeogpr.udfsgolay import udf_sgolay
from pyeogpr.udf_L8_qa import L8_cloud_qa
import pyeogpr.udfgpr

class Datacube:
    """


    pyeogpr.Datacube
    ----------------

        sensor : SENTINEL2_L1C, SENTINEL2_L2A, SENTINEL3_OLCI_L1B
            Satellite sensor to process the data with.

        biovar : Biophysical variable to process. The selected variable's map will be retrieved.
            Currently "built-in" variables available for each sensor:

            - SENTINEL2_L1C: Cab, Cm, Cw, FVC, LAI, laiCab, laiCm, laiCw
            - SENTINEL2_L2A: Cab, Cm, Cw, FVC, LAI, laiCab, laiCm, laiCw, CNC_Cab, CNC_Cprot
            - SENTINEL3_OLCI_L1B: FAPAR, FVC, LAI, LCC

        bounding_box : list
            Your region of interest. Insert bbox as list. Can be selected from https://geojson.io/
            (e.g.: [-4.55, 42.73,-4.48, 42.77])

        temporal_extent : list
            Your temporal extent to be processed. (e.g.: ["2021-01-01", "2021-12-31"])

        cloudmask : Boolean
            If "True" the Sentinel 2 cloud mask will be applied (only to S2 data), with Gaussian convolution to have
            smoother edges when masking.
            
            
    """

    def __init__(
        self,
        sensor: str,
        biovar: str,
        bounding_box: list,
        temporal_extent: list,
        cloudmask=False,
        bands= None,
    ):
        self.connection = openeo.connect(
            "https://openeo.dataspace.copernicus.eu"
        ).authenticate_oidc()
        print("""\n\n""")
        self.sensor = sensor
        if biovar[-3:] == ".py":
            self.own_model = biovar
            self.biovar = "Own_variable"
            print("own")
        else:
            self.own_model = None
            self.biovar = biovar
            print("Default")
        self.bounding_box = bounding_box
        self.temporal_extent = temporal_extent
        self.cloudmask = cloudmask
        self.spatial_extent = {
            "west": self.bounding_box[0],
            "south": self.bounding_box[1],
            "east": self.bounding_box[2],
            "north": self.bounding_box[3],
        }
        self.sensors_dict = sensors_dict
        if bands == None:
            self.bands = self.sensors_dict[self.sensor]["bandlist"]
        if bands != None:
            self.bands = bands
        self.scale_factor = None
        self.data = None
        self.masked_data = None
        self.gpr_cube = None
        self.gpr_cube_gapfilled = None
        self.models_url = "https://github.com/daviddkovacs/pyeogpr/raw/main/models/GPR_models_bulk.zip#tmp/venv"

    def construct_datacube(self, composite=None):
        """


        Parameters
        ----------
        composite : "hour","day","dekad","week","season","month","year"
            Compositing temporal interval. The resulting maps will have the following temporal steps.
            For more information: https://processes.openeo.org/#aggregate_temporal_period



        Returns
        -------
        Datacube object containing user defined parameters and temporal compositing. This will be processed into an openEO datacube.

        """

        if self.sensor not in self.sensors_dict.keys():
            print("Sensor/satellite not available as default. Using user-defined sensor.")
            scale = 0.0001
        if self.sensor in self.sensors_dict.keys():
            scale = self.sensors_dict[self.sensor]["scale_factor"]
        print(scale)
        data = (
            self.connection.load_collection(
                self.sensor,
                self.spatial_extent,
                self.temporal_extent,
                self.bands,
            )
            * scale
        )
        self.data = data
        print(self.data)
        if self.cloudmask == True and "SENTINEL2" in self.sensor:
            s2_cloudmask = self.connection.load_collection(
                "SENTINEL2_L2A", self.spatial_extent, self.temporal_extent, ["SCL"]
            )
            scl = s2_cloudmask.band("SCL")
            mask = ~((scl == 4) | (scl == 5))

            # Gaussian convolution to have a smooth edged cloud mask
            g = scipy.signal.windows.gaussian(11, std=1.6)
            kernel = np.outer(g, g)
            kernel = kernel / kernel.sum()
            mask = mask.apply_kernel(kernel)
            mask = mask > 0.1

            if composite != None:
                self.masked_data = self.data.aggregate_temporal_period(
                    composite, "mean"
                ).mask(mask)
                print(
                    f"Cloud masked, temporally composited datacube constructed: {composite} by mean values."
                )

            elif composite == None:
                self.masked_data = self.data.mask(mask)
                print("Cloud masked, datacube constructed")

        elif self.cloudmask == False and "SENTINEL2" in self.sensor:
            if composite != None:
                self.masked_data = self.data.aggregate_temporal_period(
                    composite, "mean"
                )
                print(
                    f"Temporally composited datacube constructed: {composite} by mean values. "
                )

            elif composite == None:
                self.masked_data = self.data
                print("Datacube constructed")

        elif self.cloudmask == True and "LANDSAT8_L2" in self.sensor:
            l8_qa = self.connection.load_collection(
                "LANDSAT8_L2", self.spatial_extent, self.temporal_extent, ["BQA"]
            )

            l8_cloudmask = l8_qa.apply(process=L8_cloud_qa)

            bqa = l8_cloudmask.band("BQA")
            mask = ~((bqa == 0))

            # apply kernel to mask without smoothing
            mask = mask > 0.1

            if composite != None:
                self.masked_data = self.data.aggregate_temporal_period(
                    composite, "mean"
                ).mask(mask)
                print(
                    f"Cloud masked, temporally composited datacube constructed: {composite} by mean values."
                )

            elif composite == None:
                self.masked_data = self.data.mask(mask)
                print("Cloud masked, datacube constructed")

        elif self.cloudmask == False and "LANDSAT8_L2" in self.sensor:
            if composite != None:
                self.masked_data = self.data.aggregate_temporal_period(
                    composite, "mean"
                )
                print(
                    f"Temporally composited datacube constructed: {composite} by mean values. "
                )

            elif composite == None:
                self.masked_data = self.data
                print("Datacube constructed")

        else:
            print(f"{self.sensor} can't be masked")
            self.masked_data = self.data.aggregate_temporal_period(
                composite, "mean"
            )
            print(
                f"Temporally composited datacube constructed: {composite} by mean values."
            )

    def process_map(self, gapfill=False, fileformat="nc"):
        """


        Parameters
        ----------

        gapfill : type, e.g. "Sgolay"
            To apply Savitzy Golay interpolator for cloud-induced gap 
            
        fileformat : string
                For netCDF4: "nc", for tiff: "tiff".
        
                Insert the dir of the model you developed with ARTMO. Just simply put the directory.
            

        """
        # if self.biovar not in self.sensors_dict[self.sensor]["sensor_biovar"]:
        #     print(
        #         f"'{self.biovar}' not available for this satellite/sensor"
        #     )

        if gapfill == False:
            print(f"gapfill-> {str(gapfill)}")

            if self.own_model == None:
                print(f"self.own_model {str(self.own_model)}")

                context = {"sensor": self.sensor, "biovar": self.biovar}
                self.gpr_cube = self.masked_data.apply_dimension(
                    process=udf_gpr, dimension="bands", context=context
                ).filter_bands(bands=["B02"])

                self.gpr_cube.execute_batch(
                    title=f"{self.sensor}_{self.biovar}",
                    outputfile=f"{self.sensor}_{self.biovar}.{fileformat}",
                    job_options={
                        "executor-memory": "10g",
                        "udf-dependency-archives": [self.models_url],
                    },
                )
                return

            if self.own_model != None:
                print(f"self.own_model {str(self.own_model)}")

                spec = importlib.util.spec_from_file_location("user_module", self.own_model)
                user_module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(user_module)
                # user_module = load_user_module(user_module_path)
                custom_udf = pyeogpr.udfgpr.custom_model_import(user_module)

                self.gpr_cube = self.masked_data.apply_dimension(
                    process=custom_udf, dimension="bands"
                )#.filter_bands(bands=[self.bands[0]])

                self.gpr_cube.execute_batch(
                    title="User defined product",
                    outputfile=f"user_defined_product.{fileformat}",
                    job_options={"executor-memory": "10g"},
                )
                return

        elif gapfill == True:
            print(f"gapfill-> {str(gapfill)}")

            if self.own_model == None:
                print(f"self.own_model {str(self.own_model)}")

                context = {"sensor": self.sensor, "biovar": self.biovar}

                self.gpr_cube = self.masked_data.apply_dimension(
                    process=udf_gpr, dimension="bands", context=context
                ).filter_bands(bands=["B02"])

                self.gpr_cube_gapfilled = self.gpr_cube.apply_dimension(
                    process=udf_sgolay, dimension="t"
                )

                self.gpr_cube_gapfilled.execute_batch(
                    title=f"{self.sensor} {self.biovar} gapfill->{gapfill}",
                    outputfile=f"{self.sensor}_{self.biovar}_GF.{fileformat}",
                    job_options={
                        "executor-memory": "10g",
                        "udf-dependency-archives": [self.models_url],
                    },
                )
                return

            if self.own_model != None:
                print(f"self.own_model {str(self.own_model)}")

                spec = importlib.util.spec_from_file_location("user_module", self.own_model)
                user_module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(user_module)
                # user_module = load_user_module(user_module_path)
                custom_udf = pyeogpr.udfgpr.custom_model_import(user_module)

                self.gpr_cube = self.masked_data.apply_dimension(
                    process=custom_udf, dimension="bands"
                )#.filter_bands(bands=[self.bands[0]])

                self.gpr_cube_gapfilled = self.gpr_cube.apply_dimension(
                    process=udf_sgolay, dimension="t"
                )

                self.gpr_cube.execute_batch(
                    title="User defined product",
                    outputfile=f"user_defined_product.{fileformat}",
                    job_options={"executor-memory": "10g"},
                )
                return

        else:
            raise Exception(f"'{gapfill}' is not a valid smoother")
